/* Cursed theme for gtk, based on ThinIce theme for gtk, based on
 * raster's Motif theme and the Metal theme.
 *
 * Copyright 1998 Pavel Machek <pavel@ucw.cz>, Distribute under LGPL
 * Port to GTK 2.x 2003 Johanka <johanka@ucw.cz>
 */


/* #define DEBUG 1 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <gtk/gtknotebook.h>

#include "cursed_style.h"
#include "cursed_rc_style.h"

#define STATE (state_type | FOCUS*2)
#define FOCUS ( (widget && GTK_WIDGET_HAS_FOCUS (widget)) )
#define FX (gdk_string_width(gtk_style_get_font(style),"X"))
#define FY (gdk_string_height(gtk_style_get_font(style),"X"))
#define XPOS (x)
#define YPOS (y+height/2-FY/2)

#define DETAIL(xx) ((detail) && (!strcmp(xx, detail)))
#ifndef max
#define max(x,y) ((x)>=(y)?(x):(y))
#endif
#ifndef min
#define min(x,y) ((x)<=(y)?(x):(y))
#endif

static void cursed_style_init       (CursedStyle      *style);
static void cursed_style_class_init (CursedStyleClass *klass);

static void
cursed_tab(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            const gchar *detail,
            gint x,
            gint y,
            gint width,
            gint height);


static GtkStyleClass *parent_class = NULL;

static void
sanitize_size (GdkWindow *window,
	       gint      *width,
	       gint      *height)
{
  if ((*width == -1) && (*height == -1))
    gdk_window_get_size (window, width, height);
  else if (*width == -1)
    gdk_window_get_size (window, width, NULL);
  else if (*height == -1)
    gdk_window_get_size (window, NULL, height);
}

static void
draw_hline(GtkStyle * style,
           GdkWindow * window,
           GtkStateType state_type,
           GdkRectangle * area,
           GtkWidget * widget,
           const gchar *detail,
           gint x1,
           gint x2,
           gint y)
{
  gint                thickness_light;
  gint                thickness_dark;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  thickness_light = style->ythickness / 2;
  thickness_dark = style->ythickness - thickness_light;

  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], area);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], area);
    }
  for (i = 0; i < thickness_dark; i++)
    {
      gdk_draw_line(window, style->light_gc[STATE], x2 - i - 1, y + i, x2, y + i);
      gdk_draw_line(window, style->dark_gc[STATE], x1, y + i, x2 - i - 1, y + i);
    }

  y += thickness_dark;
  for (i = 0; i < thickness_light; i++)
    {
      gdk_draw_line(window, style->dark_gc[STATE], x1, y + i, x1 + thickness_light - i - 1, y + i);
      gdk_draw_line(window, style->light_gc[STATE], x1 + thickness_light - i - 1, y + i, x2, y + i);
}
  if (area)
    { 
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], NULL);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], NULL);
    }
}

static void
draw_vline(GtkStyle * style,
           GdkWindow * window,
           GtkStateType state_type,
           GdkRectangle * area,
           GtkWidget * widget,
           const gchar *detail,
           gint y1,
           gint y2,
           gint x)
{
  gint                thickness_light;
  gint                thickness_dark;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  thickness_light = style->xthickness / 2;
  thickness_dark = style->xthickness - thickness_light;

  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], area);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], area);
    }
  for (i = 0; i < thickness_dark; i++)
    {
      gdk_draw_line(window, style->light_gc[STATE], x + i, y2 - i - 1, x + i, y2);
      gdk_draw_line(window, style->dark_gc[STATE], x + i, y1, x + i, y2 - i - 1);
    }

  x += thickness_dark;
  for (i = 0; i < thickness_light; i++)
    {
      gdk_draw_line(window, style->dark_gc[STATE], x + i, y1, x + i, y1 + thickness_light - i);
      gdk_draw_line(window, style->light_gc[STATE], x + i, y1 + thickness_light - i, x + i, y2);
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], NULL);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], NULL);
    }
}

static void
draw_shadow(GtkStyle     *style,
	    GdkWindow    *window,
	    GtkStateType  state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle *area,
	    GtkWidget    *widget,
	    const gchar  *detail,
	    gint          x,
	    gint          y,
	    gint          width,
	    gint          height)
{
  GdkGC *gc1 = NULL;
  GdkGC *gc2 = NULL;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      return;
    case GTK_SHADOW_IN:
    case GTK_SHADOW_ETCHED_IN:
      gc1 = style->light_gc[STATE];
      gc2 = style->dark_gc[STATE];
      break;
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = style->dark_gc[STATE];
      gc2 = style->light_gc[STATE];
      break;
    }

  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      if ((shadow_type == GTK_SHADOW_IN) ||
	  (shadow_type == GTK_SHADOW_OUT))
	{
	  gdk_gc_set_clip_rectangle(style->black_gc, area);
	  gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
	}
    }
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      break;

    case GTK_SHADOW_ETCHED_IN:
	gdk_draw_rectangle(window, gc2, FALSE, x, y, width-1, height-1);
/*	gdk_draw_rectangle(window, gc2, FALSE, x+1, y+1, width-3, height-3); */

	break;
    case GTK_SHADOW_IN:
	gdk_draw_line(window, gc2,
		      x, y, x + width-1, y);
	gdk_draw_line(window, gc2,
		      x, y, x, y + height-1);
	
	gdk_draw_line(window, gc1,
		      x, y + height-1, x + width-1, y + height-1);
	gdk_draw_line(window, gc1,
		      x + width-1, y, x + width-1, y + height-1);
	break;
	
    case GTK_SHADOW_ETCHED_OUT:
	gdk_draw_rectangle(window, gc2, FALSE, x, y, width-1, height-1);

	break;
    case GTK_SHADOW_OUT:
	gdk_draw_line(window, gc2,
		      x, y, x + width-1, y);
	gdk_draw_line(window, gc2,
		      x, y, x, y + height-1);
	
	gdk_draw_line(window, gc1,
		      x, y + height-1, x + width-1, y + height-1);
	gdk_draw_line(window, gc1,
		      x + width-1, y, x + width-1, y + height-1);
	
	break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      if ((shadow_type == GTK_SHADOW_IN) ||
	  (shadow_type == GTK_SHADOW_OUT))
	{
	  gdk_gc_set_clip_rectangle(style->black_gc, NULL);
	  gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
	}
    }
}

static void
draw_polygon(GtkStyle * style,    
             GdkWindow * window,
             GtkStateType state_type,
             GtkShadowType shadow_type,
             GdkRectangle * area,
             GtkWidget * widget,
             const gchar * detail,
             GdkPoint * points,
             gint npoints,
             gint fill)
{
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif /* M_PI */
#ifndef M_PI_4
#define M_PI_4  0.78539816339744830962
#endif /* M_PI_4 */

  static const gdouble pi_over_4 = M_PI_4;
  static const gdouble pi_3_over_4 = M_PI_4 * 3;

  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;
  GdkGC              *gc4;
  gdouble             angle;
  gint                xadjust;
  gint                yadjust;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  g_return_if_fail(points != NULL);

  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gc1 = style->light_gc[STATE];
      gc2 = style->dark_gc[STATE];
      gc3 = style->light_gc[STATE];
      gc4 = style->dark_gc[STATE];
      break;
    case GTK_SHADOW_ETCHED_IN:
      gc1 = style->light_gc[STATE];
      gc2 = style->dark_gc[STATE];
      gc3 = style->dark_gc[STATE];
      gc4 = style->light_gc[STATE];
      break;
    case GTK_SHADOW_OUT:
      gc1 = style->dark_gc[STATE];
      gc2 = style->light_gc[STATE];
      gc3 = style->dark_gc[STATE];
      gc4 = style->light_gc[STATE];
      break;
    case GTK_SHADOW_ETCHED_OUT:
      gc1 = style->dark_gc[STATE];
      gc2 = style->light_gc[STATE];
      gc3 = style->light_gc[STATE];
      gc4 = style->dark_gc[STATE];
      break;
    default:
      return;
    }
  

  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, area);
      gdk_gc_set_clip_rectangle(gc2, area);
      gdk_gc_set_clip_rectangle(gc3, area);
      gdk_gc_set_clip_rectangle(gc4, area);
    }
  if (fill)
    gdk_draw_polygon(window, style->bg_gc[STATE], TRUE, points, npoints);
    npoints--;

    for (i = 0; i < npoints; i++)
      {
        if ((points[i].x == points[i + 1].x) &&
           (points[i].y == points[i + 1].y))
          {
            angle = 0;
          }
        else
          {
            angle = atan2(points[i + 1].y - points[i].y,
            points[i + 1].x - points[i].x);
          }

        if ((angle > -pi_3_over_4) && (angle < pi_over_4))
          {
          if (angle > -pi_over_4)
            {
              xadjust = 0;
              yadjust = 1;
            }
          else
	    {
	      xadjust = 1;
	      yadjust = 0;
	    }

          gdk_draw_line(window, gc1,
	  points[i].x - xadjust, points[i].y - yadjust,
	  points[i + 1].x - xadjust, points[i + 1].y - yadjust);
          gdk_draw_line(window, gc3,
	  points[i].x, points[i].y,
	  points[i + 1].x, points[i + 1].y);
        }
      else
        {
          if ((angle < -pi_3_over_4) || (angle > pi_3_over_4))
            {
              xadjust = 0;
              yadjust = 1;
            }
          else
            {
              xadjust = 1;
              yadjust = 0;
            }

          gdk_draw_line(window, gc4,
          points[i].x + xadjust, points[i].y + yadjust,
          points[i + 1].x + xadjust, points[i + 1].y + yadjust);
          gdk_draw_line(window, gc2,
          points[i].x, points[i].y,
          points[i + 1].x, points[i + 1].y);
       }
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(gc1, NULL);
      gdk_gc_set_clip_rectangle(gc2, NULL);
      gdk_gc_set_clip_rectangle(gc3, NULL);
      gdk_gc_set_clip_rectangle(gc4, NULL);
   }
}


static void
draw_arrow(GtkStyle * style,       
           GdkWindow * window,
           GtkStateType state_type,
           GtkShadowType shadow_type,
           GdkRectangle * area,
           GtkWidget * widget,
           const gchar *detail,
           GtkArrowType arrow_type,
           gint fill, gint x, gint y, gint width, gint height)
{
  GdkWChar up = 0x2191;
  GdkWChar down = 0x2193;
	
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);


  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  switch (arrow_type) {
    case 0:
      gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], x+(width-FX)/2, y+(height-FY)/2, &up, 1); 
      break;
    case 1:
      gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], x+(width-FX)/2, y+(height-FY)/2, &down, 1); 
      break;  
    case 2:
      gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x+(width-FX)/2, y+(height-FY)/2, "<"); 
      break;  
    case 3:
      gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x+(width-FX)/2, y+(height-FY)/2, ">"); 
      break;
    default: 
      break;
  }
}


static void
draw_diamond(GtkStyle * style,
             GdkWindow * window,
             GtkStateType state_type,
             GtkShadowType shadow_type,
             GdkRectangle * area,
             GtkWidget * widget,
             const gchar *detail,
             gint x,
             gint y,
             gint width,
             gint height)
{
  gint                half_width;
  gint                half_height;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  half_width = width / 2;
  half_height = height / 2;

  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], area);
      gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], area);
      gdk_gc_set_clip_rectangle(style->black_gc, area);
    }
  switch (shadow_type)
    {
    case GTK_SHADOW_IN:
      gdk_draw_line(window, style->light_gc[STATE],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[STATE],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->dark_gc[STATE],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    case GTK_SHADOW_OUT:
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + 2, y + half_height,
		    x + half_width, y + height - 2);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y + height - 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + 1, y + half_height,
		    x + half_width, y + height - 1);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y + height - 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x, y + half_height,
		    x + half_width, y + height);
      gdk_draw_line(window, style->dark_gc[STATE],
		    x + half_width, y + height,
		    x + width, y + half_height);

      gdk_draw_line(window, style->light_gc[STATE],
		    x + 2, y + half_height,
		    x + half_width, y + 2);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y + 2,
		    x + width - 2, y + half_height);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + 1, y + half_height,
		    x + half_width, y + 1);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y + 1,
		    x + width - 1, y + half_height);
      gdk_draw_line(window, style->light_gc[STATE],
		    x, y + half_height,
		    x + half_width, y);
      gdk_draw_line(window, style->light_gc[STATE],
		    x + half_width, y,
		    x + width, y + half_height);
      break;
    default:
      break;
    }
  if (area)
    {
      gdk_gc_set_clip_rectangle(style->light_gc[STATE], NULL);
      gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
      gdk_gc_set_clip_rectangle(style->dark_gc[STATE], NULL);
      gdk_gc_set_clip_rectangle(style->black_gc, NULL);
    }
}

static void
draw_box(GtkStyle * style,
         GdkWindow * window,
         GtkStateType state_type,
         GtkShadowType shadow_type,
         GdkRectangle * area,
         GtkWidget * widget,
         const gchar *detail,
         gint x,
         gint y,
         gint width,
         gint height)
{
  GdkGC *light_gc, *dark_gc;
  GtkOrientation      orientation;
  GdkWChar horiz = 0x2500;
  GdkWChar vert = 0x2502;
  GdkWChar down = 0x2193;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);


  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  light_gc = style->light_gc[STATE];
  dark_gc = style->dark_gc[STATE];

  orientation = GTK_ORIENTATION_HORIZONTAL;
  if (height > width)
    orientation = GTK_ORIENTATION_VERTICAL;

#ifdef DEBUG
     printf("%p %s %i %i state_type = %d (%d)\n", detail, detail, width, height, state_type, orientation);
#endif

  if (DETAIL("through")) {
      if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
      }
      gdk_draw_rectangle(window, style->bg_gc[STATE], TRUE,
              x, y, width, height);
      if (area) {
          gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
      }
      gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
              detail, x, y, width, height);
  } else if (DETAIL("slider")) {
        int i;
/*        ThemeStyleData *data = style->engine_data; */

	if (orientation == GTK_ORIENTATION_HORIZONTAL) {
	  for (i=x; i<x+width; i+=8) {
	    gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], i, y+FY, &horiz, 1);
	  }
	} else {
	  for (i=y; i<y+height; i+=8) {
	    gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], x, i+FY, &vert, 1);
	  }
	}
  } else if (DETAIL("optionmenutab")) {
    gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], XPOS, YPOS, &down, 1);
    } else if (DETAIL("buttondefault")) {
      /* I don't want no background on default buttons..
         Let's add that cute triangle (see below) instead... */
    } else if (DETAIL("button")) {
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
        }

	gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x, YPOS, "[");
	gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x+width-FX, YPOS, "]");

        /* Paint a triangle here instead of in "buttondefault"
           which is drawn _behind_ the current button */
        if (GTK_WIDGET_HAS_DEFAULT (widget)) {              /* FIXME */
	    gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x+FX, YPOS, "#");
        }
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
        }
    } else if (DETAIL("bar")) {
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[GTK_STATE_SELECTED],
                TRUE, x, y, width, height);
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], NULL);
        }
    } else if (DETAIL("handlebox_bin")) {
      /* Do we want anything behind the handlebox-buttons? */
        /*
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[GTK_STATE_SELECTED],
                TRUE, x + 9, y + 1, width - 2, height - 2);

        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[GTK_STATE_SELECTED], NULL);
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
         */
    } else if (DETAIL("menubar")) {
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
        }
        gdk_draw_rectangle(window, style->bg_gc[STATE], TRUE,
                x, y, width, height);
        if (area) {
            gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
        }
    } else if (DETAIL("tab")) {
        cursed_tab(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    } else {
        if ((!style->bg_pixmap[state_type]) ||
                (GDK_IS_PIXMAP(window))) {
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
            }
            gdk_draw_rectangle(window, style->bg_gc[STATE], TRUE,
                    x, y, width, height);
            if (area) {
                gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
            }
        } else {
        }
        gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
                detail, x, y, width, height);
    }
}
  
static void
draw_check(GtkStyle * style,
           GdkWindow * window,
           GtkStateType state_type,
           GtkShadowType shadow_type,
           GdkRectangle * area,
           GtkWidget * widget,
           const gchar *detail,
           gint x,
           gint y,
           gint width,
           gint height)
{
  char buf[]="[!]";

  buf[1] = (shadow_type == GTK_SHADOW_IN)?'x':' ';
  gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x, YPOS, buf); 
}

/* Thanks to Evan Lawrence */
static void
draw_option(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            const gchar *detail,
            gint x,
            gint y,
            gint width,
            gint height)
{
  char buf[]="(!)";
  buf[1] = (shadow_type == GTK_SHADOW_IN)?'o':' ';
  gdk_draw_string(window, gtk_style_get_font(style), style->fg_gc[STATE], x, YPOS, buf);
}


static void 
draw_tab(GtkStyle * style,
         GdkWindow * window,
         GtkStateType state_type,
         GtkShadowType shadow_type,
         GdkRectangle * area,
         GtkWidget * widget,
         const gchar *detail,
         gint x,
         gint y,
         gint width,
         gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
                x, y, width, height);
}



static void
draw_shadow_gap(GtkStyle * style,
                GdkWindow * window,
                GtkStateType state_type,
                GtkShadowType shadow_type,
                GdkRectangle * area,
                GtkWidget * widget,
                const gchar *detail,
                gint x,
                gint y,
                gint width,
                gint height,
                GtkPositionType gap_side,
                gint gap_x,
                gint gap_width)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_shadow(style, window, state_type, shadow_type, area, widget, detail,
     x, y, width, height);

  switch (gap_side)
    {
      case GTK_POS_TOP:
        rect.x = x + gap_x;
        rect.y = y;
        rect.width = gap_width;
        rect.height = 2;
      break;
      case GTK_POS_BOTTOM:
        rect.x = x + gap_x;
        rect.y = y + height - 2;
        rect.width = gap_width;
        rect.height = 2;
      break;
												          case GTK_POS_LEFT:
        rect.x = x;
        rect.y = y + gap_x;
        rect.width = 2;
        rect.height = gap_width;
      break;
      case GTK_POS_RIGHT:
        rect.x = x + width - 2;
        rect.y = y + gap_x;
        rect.width = 2;
        rect.height = gap_width;
      break;
   }
  
}


static void
draw_box_gap(GtkStyle * style,
             GdkWindow * window,
             GtkStateType state_type,
             GtkShadowType shadow_type,
             GdkRectangle * area,
             GtkWidget * widget,
             const gchar *detail,
             gint x,
             gint y,
             gint width,
             gint height,
             GtkPositionType gap_side,
             gint gap_x,
             gint gap_width)
{
    sanitize_size (window, &width, &height);

    gtk_style_apply_default_background(style, window,
            widget && !GTK_WIDGET_NO_WINDOW(widget),
            state_type, area,
            x, y, width, height);
    draw_shadow_gap (style, window, state_type, shadow_type,
            area, widget, detail, x, y, width, height,
            gap_side, gap_x, gap_width);
}


static void
draw_extension(GtkStyle * style,
               GdkWindow * window,
               GtkStateType state_type,
               GtkShadowType shadow_type,
               GdkRectangle * area,
               GtkWidget * widget,
               const gchar *detail,
               gint x,
               gint y,
               gint width,
               gint height,
               GtkPositionType gap_side)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  sanitize_size (window, &width, &height);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
                x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + style->xthickness;
      rect.y = y;
      rect.width = width - style->xthickness * 2;
      rect.height = style->ythickness;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + style->xthickness;
      rect.y = y + height - style->ythickness;
      rect.width = width - style->xthickness * 2;
      rect.height = style->ythickness;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + style->ythickness;
      rect.width = style->xthickness;
      rect.height = height - style->ythickness * 2;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - style->xthickness;
      rect.y = y + style->ythickness;
      rect.width = style->xthickness;
      rect.height = height - style->ythickness * 2;
      break;
    }

}

static void
draw_slider(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            const gchar *detail,
            gint x,
            gint y,
            gint width,
            gint height,
            GtkOrientation orientation)
{
  GdkWChar buf;
  int i;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  buf = 0x2592;

  
  if (orientation == GTK_ORIENTATION_HORIZONTAL) {
    for (i = 0; i < width; i++) 
      gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], x+i, YPOS, &buf, 1);
  } else {
    for (i = 0; i < height; i++)	  
    gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], XPOS, y+i, &buf, 1);
  }

  
}

static void
draw_focus (GtkStyle      *style,
            GdkWindow     *window,
            GtkStateType   state_type,
            GdkRectangle  *area,
            GtkWidget     *widget,
            const gchar   *detail,
            gint           x,
            gint           y,
            gint           width,
            gint           height)
{
  /* GTK+ default draw_focus() would fill a rectangle over the text
   * with color fg[NORMAL] clearing out any previous text there.
   */
}

static void
draw_handle(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            const gchar *detail,
            gint x,
            gint y,
            gint width,
            gint height,
            GtkOrientation orientation)
{
  int i;
  const GdkWChar WChar_BOX_DRAWINGS_LIGHT_HORIZONTAL = 0x2500;
  const GdkWChar WChar_BOX_DRAWINGS_LIGHT_VERTICAL = 0x2502;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

    if ((width == -1) && (height == -1))
      gdk_window_get_size(window, &width, &height);
    else if (width == -1)
      gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
      gdk_window_get_size(window, NULL, &height);

  if (orientation == GTK_ORIENTATION_HORIZONTAL) {
    for (i = 0; i < width; i++) 
      gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], x+i, YPOS,
		       &WChar_BOX_DRAWINGS_LIGHT_HORIZONTAL, 1);
  } else {
    for (i = 0; i < height; i++)	  
      gdk_draw_text_wc(window, gtk_style_get_font(style), style->fg_gc[STATE], XPOS, y+i,
		       &WChar_BOX_DRAWINGS_LIGHT_VERTICAL, 1);
  }
}

static void
cursed_tab(GtkStyle * style,
            GdkWindow * window,
            GtkStateType state_type,
            GtkShadowType shadow_type,
            GdkRectangle * area,
            GtkWidget * widget,
            const gchar *detail,
            gint x,
            gint y,
            gint width,
            gint height)
{
  GtkNotebook *notebook;
  GdkGC *lightgc, *darkgc;
  int orientation;


  notebook = GTK_NOTEBOOK(widget);
  orientation = notebook->tab_pos;

  lightgc = style->light_gc[STATE];
  darkgc = style->dark_gc[STATE];

  if ((!style->bg_pixmap[state_type]) ||
    (GDK_IS_PIXMAP(window))) {
    if (area) {
      gdk_gc_set_clip_rectangle(style->bg_gc[STATE], area);
    }
    gdk_draw_rectangle(window, style->bg_gc[STATE], TRUE,
      x, y, width, height);
    if (area) {
      gdk_gc_set_clip_rectangle(style->bg_gc[STATE], NULL);
    }
  } else {
  }
  if (area) {
    gdk_gc_set_clip_rectangle(style->dark_gc[STATE], area);
    gdk_gc_set_clip_rectangle(style->light_gc[STATE], area);
  }
  switch(orientation) {
    case GTK_POS_TOP:
      gdk_draw_line(window, lightgc,
        x, y + height - 1, x, y);
      gdk_draw_line(window, lightgc,
        x, y, x + width - 1, y);
      gdk_draw_line(window, darkgc,
        x + width - 1, y, x + width - 1, y + height - 1);
    break;
    case GTK_POS_BOTTOM:
      gdk_draw_line(window, lightgc,
        x, y, x, y + height - 1);
      gdk_draw_line(window, darkgc,
        x, y + height - 1, x + width - 1, y + height - 1);
      gdk_draw_line(window, darkgc,
        x + width - 1, y + height - 1, x + width - 1, y);
    break;
    case GTK_POS_LEFT:
      gdk_draw_line(window, lightgc,
        x, y + height - 1, x, y);
      gdk_draw_line(window, lightgc,
        x, y, x + width - 1, y);
      gdk_draw_line(window, darkgc,
        x, y + height - 1, x + width - 1, y + height - 1);
    break;
    case GTK_POS_RIGHT:
      gdk_draw_line(window, lightgc,
        x, y, x + width - 1, y);
      gdk_draw_line(window, darkgc,
        x + width - 1, y, x + width - 1, y + height - 1);
      gdk_draw_line(window, darkgc,
        x, y + height - 1, x + width - 1, y + height - 1);
      break;
   }
   if (area) {
     gdk_gc_set_clip_rectangle(style->dark_gc[STATE], NULL);
     gdk_gc_set_clip_rectangle(style->light_gc[STATE], NULL);
   }
}


  
GType cursed_type_style = 0;

void
cursed_style_register_type (GTypeModule *module)
{
  static const GTypeInfo object_info =
  {
    sizeof (CursedStyleClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) cursed_style_class_init,
    NULL,           /* class_finalize */
    NULL,           /* class_data */
    sizeof (CursedStyle),
    0,              /* n_preallocs */
    (GInstanceInitFunc) cursed_style_init,
  };
  
  cursed_type_style = g_type_module_register_type (module,
						    GTK_TYPE_STYLE,
						    "CursedStyle",
						    &object_info, 0);
}

static void
cursed_style_init (CursedStyle *style)
{
}

static void
cursed_style_class_init (CursedStyleClass *klass)
{
  GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  style_class->draw_hline = draw_hline;
  style_class->draw_vline = draw_vline;
  style_class->draw_shadow = draw_shadow;
  style_class->draw_polygon = draw_polygon;
  style_class->draw_arrow = draw_arrow;
  style_class->draw_diamond = draw_diamond;
  style_class->draw_box = draw_box;
  style_class->draw_check = draw_check;
  style_class->draw_option = draw_option;
  style_class->draw_tab = draw_tab;
  style_class->draw_shadow_gap = draw_shadow_gap;
  style_class->draw_box_gap = draw_box_gap;
  style_class->draw_extension = draw_extension;
  style_class->draw_focus = draw_focus;
  style_class->draw_slider = draw_slider;
  style_class->draw_handle = draw_handle;
}



